/*
 * File:		mcf5206_hi.c
 * Purpose:		Higher level processing for MCF5206 specifics.
 *
 * Notes:
 *
 * Date:		7-3-96
 *
 * Modifications:
 *
 */

#include "uif/dbug.h"

/*
 * CPU name string and source code version for display at boot-up.
 */
const char CPUSTR[] = "ColdFire MCF5206";
const int  CPUVER   = 1;

/*
 * This defines the data structure which will hold the values
 * of the registers of the user task.
 */
REGISTERS context;

/*
 * This structure contains info about the registers.  The name
 * is given so that users can type it.
 */
static const reg_entry registers[] =
{
	{ "D0", (void *)&context.d0, 32 },
	{ "D1", (void *)&context.d1, 32 },
	{ "D2", (void *)&context.d2, 32 },
	{ "D3", (void *)&context.d3, 32 },
	{ "D4", (void *)&context.d4, 32 },
	{ "D5", (void *)&context.d5, 32 },
	{ "D6", (void *)&context.d6, 32 },
	{ "D7", (void *)&context.d7, 32 },
	{ "A0", (void *)&context.a0, 32 },
	{ "A1", (void *)&context.a1, 32 },
	{ "A2", (void *)&context.a2, 32 },
	{ "A3", (void *)&context.a3, 32 },
	{ "A4", (void *)&context.a4, 32 },
	{ "A5", (void *)&context.a5, 32 },
	{ "A6", (void *)&context.a6, 32 },
	{ "A7", (void *)&context.a7, 32 },
	{ "SP", (void *)&context.a7, 32 },	/* SP is A7 */
	{ "PC", (void *)&context.pc, 32 },
	{ "SR", (void *)&context.sr, 16 },
} ;
#define	REGTABLESIZE (sizeof(registers) / sizeof(reg_entry))

static const char INVREG[] = "Error: Invalid Register: %s\n";
static const char INVMOD[] = "Error: Invalid Module: %s\n";

/* a value of 0 means do RTE, a non-zero value means back to monitor */
#define EXEC_RTE 0
#define EXEC_DBUG 1

int 	cpu_trace_count;
ADDRESS trace_thru;
ADDRESS step_over;


/*************************************************/
void
cpu_init (void)
{
	trace_thru = NULL;
	step_over = NULL;
	cpu_trace_count = 0;

	/* init user regs */
	context.d0 = 0;
	context.d1 = 0;
	context.d2 = 0;
	context.d3 = 0;
	context.d4 = 0;
	context.d5 = 0;
	context.d6 = 0;
	context.d7 = 0;
	context.a0 = 0;
	context.a1 = 0;
	context.a2 = 0;
	context.a3 = 0;
	context.a4 = 0;
	context.a5 = 0;
	context.a6 = 0;
	context.a7 = 0;	/* filled in by board */
	context.pc = 0;
	context.sr = 0x2000;
}


/*************************************************/
int
cpu_find_register (char *reg)
{
	/*
	 * This routine searches the register table for the
	 * register named `reg'.  If it is found, then the
	 * index needed to access that register in the table
	 * is returned, otherwise -1 is returned.
	 */
	int index;

	for (index = 0; index < REGTABLESIZE; index++)
	{
		if (strcasecmp(registers[index].reg_name,reg) == 0)
			return index;
	}
	return -1;
}


/*************************************************/
void
cpu_set_register(char *reg, NATURAL32 value)
{
	int index;
	if ((index = cpu_find_register(reg)) != -1)
	{
		switch (registers[index].reg_size)
		{
			case SIZE_16:
				*(NATURAL16 *)registers[index].reg_offset = (NATURAL16)value;
				break;
			case SIZE_32:
			default:
				*(NATURAL32 *)registers[index].reg_offset = (NATURAL32)value;
				break;
		}
	}
	else
	{
		printf(INVREG,reg);
	}
}

/*************************************************/
void
cpu_set_program_counter(ADDRESS address)
{
	context.pc = (NATURAL32)address;
}


/*************************************************/
ADDRESS
cpu_get_program_counter(void)
{
	return (ADDRESS)context.pc;
}

/*************************************************/
int
cpu_parse_size (char *arg)
{
	int i, size = SIZE_16;

	for (i = 0; arg[i] != '\0'; i++)
	{
		if (arg[i] == '.')
		{
			switch (arg[i+1])
			{
				case 'b':
				case 'B':
					size = SIZE_8;
					break;
				case 'w':
				case 'W':
					size = SIZE_16;
					break;
				case 'l':
				case 'L':
					size = SIZE_32;
					break;
				default:
					size = SIZE_16;
					break;
			}
			break;
		}
	}
	return size;
}


/*************************************************/
static void
cpu_display_sr (NATURAL16 sr)
{
	char sr_buffer[17];
	register int i;
	INTEGER32 mask = 0x00008000;
	INTEGER32 result;

	for (i = 0; i < 16; i++)
	{
		result = mask & (long)sr;
		mask = mask >> 1;

		switch (i)		/* going from left to right, don't be fooled */
		{
			case 0:
				if (result)
					sr_buffer[i] = 'T';
				else
					sr_buffer[i] = 't';
				break;
			case 1:
			case 4:
			case 8:
			case 9:
			case 10:
					sr_buffer[i] = '.';
					break;
			case 2:
				if (result)
					sr_buffer[i] = 'S';
				else
					sr_buffer[i] = 's';
				break;
			case 3:
				if (result)
					sr_buffer[i] = 'M';
				else
					sr_buffer[i] = 'm';
				break;
			case 5:
			case 6:
			case 7:
				if (result)
					sr_buffer[i] = '1';
				else
					sr_buffer[i] = '0';
				break;
			case 11:
				if (result)
					sr_buffer[i] = 'X';
				else
					sr_buffer[i] = 'x';
				break;
			case 12:
				if (result)
					sr_buffer[i] = 'N';
				else
					sr_buffer[i] = 'n';
				break;
			case 13:
				if (result)
					sr_buffer[i] = 'Z';
				else
					sr_buffer[i] = 'z';
				break;
			case 14:
				if (result)
					sr_buffer[i] = 'V';
				else
					sr_buffer[i] = 'v';
				break;
			case 15:
				if (result)
					sr_buffer[i] = 'C';
				else
					sr_buffer[i] = 'c';
				break;
			default:
				break;
		}
	}
	sr_buffer[i] = 0;
	printf("%s",sr_buffer);
}


/*************************************************/
void
cpu_display_register(char *reg)
{
	int index;

	if (reg == NULL)
	{
		printf("PC: %08X ",  context.pc);
		printf("SR: %04X [", context.sr);
		cpu_display_sr(context.sr);
		printf("]\n");
		printf("An: %08X %08X %08X %08X %08X %08X %08X %08X\n",
			context.a0,
			context.a1,
			context.a2,
			context.a3,
			context.a4,
			context.a5,
			context.a6,
			context.a7  );
			
		printf("Dn: %08X %08X %08X %08X %08X %08X %08X %08X\n",
			context.d0,
			context.d1,
			context.d2,
			context.d3,
			context.d4,
			context.d5,
			context.d6,
			context.d7  );
	}
	else
	{
		/* display specific reg */
		if ((index = cpu_find_register(reg)) != -1)
		{
			switch (registers[index].reg_size)
			{
				case SIZE_16:
					printf("%s:  %04X\n",
						registers[index].reg_name,
						*(NATURAL16 *)registers[index].reg_offset );
					break;
				case SIZE_32:
				default:
					printf("%s:  %08X\n",
						registers[index].reg_name,
						*(NATURAL32 *)registers[index].reg_offset );
			}
		}
		else
		{
			printf(INVREG,reg);
		}
	}
}

/*************************************************/
int
cpu_invalid_instr_addr(ADDRESS address)
{
	/* ColdFire requires instructions be on word boundary */
	return ((NATURAL32)address & (NATURAL32)0x00000001);
}

/*************************************************/
void
cpu_run_user_code (int trace)
{
	/* if `trace' is TRUE, then don't put in breakpoints */
	/* otherwise install breakpoints, and then execute   */

	if (context.a7 < 0x0400)	/* vector table */
	{
		printf("Error:  Bad Stack Pointer\n");
		return;
	}

	if (trace)
	{
		/* turn tracing on */
		context.sr |= MCF5206_SR_T;
	}
	else
	{
	    if (breakpoint_install((ADDRESS)context.pc))
		{
			/* about to execute at a breakpoint */
			/* trace thru this breakpoint rather breaking */
			trace_thru = (ADDRESS)context.pc;
			/* turn tracing on */
			context.sr |= MCF5206_SR_T;
		}
	}
	asm_switch_context(&context);
}


/*************************************************/
void
cpu_write_data (ADDRESS address, int size, int data)
{
	/* achieve proper alignment and write data */
	switch (size)
	{
		case SIZE_8:
			*((BYTE *)address) = (BYTE)data;
			break;
		case SIZE_16:
			address = (ADDRESS)((int)address & 0xFFFFFFFE);
			*((WORD *)address) = (WORD)data;
			break;
		case SIZE_32:
			address = (ADDRESS)((int)address & 0xFFFFFFFC);
			*((LONG *)address) = (LONG)data;
			break;
		default:
			break;
	}
}

/*************************************************/
int
cpu_read_data (ADDRESS address, int size)
{
	switch (size)
	{
		case SIZE_8:
			return *((BYTE *)address);
			break;
		case SIZE_16:
			return *((WORD *)address);
			break;
		case SIZE_32:
			return *((LONG *)address);
			break;
		default:
			return 0;
			break;
	}
}


/*************************************************/
ADDRESS
cpu_align_address (ADDRESS address, int size)
{
	switch (size)
	{
		case SIZE_16:
			return (address & 0xFFFFFFFE);
			break;
		case SIZE_32:
			return (address & 0xFFFFFFFC);
			break;
		case SIZE_8:
		default:
			return (address);
			break;
	}
}

/*************************************************/
static void
display_fs (STACK_FRAME *framep)
{
	int fault = SF_FS(framep);

	printf(" FS=%X, ",fault);

	switch (fault)
	{
		case 0x04:
			printf("Physical bus error on instruction fetch");
			break;
		case 0x08:
			printf("Physical bus error on operand write");
			break;
		case 0x09:
			printf("Attempted write to write-protected space");
			break;
		case 0x0C:
			printf("Physical bus error on operand read");
			break;
		default:
			printf("Reserved");
			break;
	}

	printf("\n");
}

/*************************************************/
int
cpu_handler (STACK_FRAME *framep)
{
	/*
	 * This is the exception handler for all defined exceptions.  Most
	 * exceptions do nothing, but some of the more important ones are
	 * handled to some extent.  The Autovector handlers should eventually
	 * be handled separately so that they can efficiently handle
	 * peripheral and I/O services.
	 */
	int ispdelta;
	int cpu_handler_flag;
	int user_brkpnt, user_triggered;

	/*
	 * Default action for most exceptions is to break to monitor.
	 */
	cpu_handler_flag = EXEC_DBUG;

	/*
	 * Update context.a7 to point to prior to stack frame.
	 */
	switch (SF_FORMAT(framep))
	{
		case 4:
			ispdelta = 8;
			break;
		case 5:
			ispdelta = 9;
			break;
		case 6:
			ispdelta = 10;
			break;
		case 7:
			ispdelta = 11;
			break;
		default:
			printf("\nIllegal stack type!\n");
			ispdelta = 0;
			break;
	}
	context.a7 = context.a7 + ispdelta;

	user_brkpnt = breakpoint_deinstall((ADDRESS)context.pc, &user_triggered);

	switch (SF_VECTOR(framep))
	{
		case 2:
			printf("Access Error: ");
			display_fs(framep);
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 3:
			printf("Address Error: ");
			display_fs(framep);
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 4:
			if (!user_brkpnt)
			{
				printf("Illegal instruction at %#08X\n",context.pc);
				cpu_display_register(NULL);
				cpu_disasm((ADDRESS)context.pc,1);
			}
			else
			{
				if ((ADDRESS)step_over == (ADDRESS)context.pc)
				{
					/* clean up is at end of handler */
					cpu_display_register(NULL);
					cpu_disasm((ADDRESS)context.pc,1);
					break;
				}
				if (user_triggered)
				{
					printf("Breakpoint encountered at %#08X\n",context.pc);
					cpu_display_register(NULL);
					cpu_disasm((ADDRESS)context.pc,1);
				}
				/* else reinstall breakpoints and continue */
				/* execution of the task...we will return  */
				/* to task from here, not write out regs and RTE */
				else
				{	
			    	if (breakpoint_install((ADDRESS)context.pc))
					{
						/* about to execute at a breakpoint */
						/* trace thru this breakpoint rather breaking */
						trace_thru = (ADDRESS)context.pc;
						/* turn tracing on */
						SF_SR(framep) |= MCF5206_SR_T;
						/* readjust stack pointer for RTE */
						context.a7 -= ispdelta;
						cpu_handler_flag = EXEC_RTE;
					}
				}
			}
			break;
		case 8:
			printf("Privilege violation.\n");
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 9:
			if ((int)trace_thru)
			{
				/* place breakpoint at trace_thru */
				cpu_write_data(trace_thru,16,(int)ILLEGAL);
				/* turn off tracing */
				context.sr &= (~MCF5206_SR_T);
				/* nullify trace_thru and continue execution */
				trace_thru = 0;
				cpu_handler_flag = EXEC_RTE;
				/* adjust user stack back so RTE frame correct */
				context.a7 = context.a7 - ispdelta;
				/* since not recreating a new stack frame, must */
				/* turn off trace bit in SR already on stack.   */
				SF_SR(framep) &= (~MCF5206_SR_T);
				break;
			}

			/* each time through, decrement cpu_trace_count */
			if (--cpu_trace_count > 0)
			{
				context.a7 -= ispdelta;
				cpu_handler_flag = EXEC_RTE;
			}
			else
			{
				/* since not recreating a new stack frame, must */
				/* turn off trace bit in SR already on stack.   */
				SF_SR(framep) &= (~MCF5206_SR_T);
				/* turn off tracing */
				context.sr &= (~MCF5206_SR_T);
				cpu_handler_flag = EXEC_DBUG;
				cpu_trace_count = 0;
			}
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 10:
			printf("Unimplemented A-Line Instruction\n");
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 11:
			printf("Unimplemented F-Line Instruction\n");
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 12:
			printf("Debug Interrupt\n");
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 14:
			printf("Format Error\n");
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 15:
			printf("Unitialized Interrupt\n");
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 24:
			printf("Spurious Interrupt\n");
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 25:
		case 26:
		case 27:
		case 28:
		case 29:
		case 30:
		case 31:
			printf("Autovector interrupt level %d\n",SF_VECTOR(framep) - 24);
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 32:
		case 33:
		case 34:
		case 35:
		case 36:
		case 37:
		case 38:
		case 39:
		case 40:
		case 41:
		case 42:
		case 43:
		case 44:
		case 45:
		case 46:
		case 47:
			printf("TRAP #%d\n",SF_VECTOR(framep) - 32);
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		case 5:
		case 6:
		case 7:
		case 13:
		case 16:
		case 17:
		case 18:
		case 19:
		case 20:
		case 21:
		case 22:
		case 23:
		case 48:
		case 49:
		case 50:
		case 51:
		case 52:
		case 53:
		case 54:
		case 55:
		case 56:
		case 57:
		case 58:
		case 59:
		case 60:
		case 61:
		case 62:
		case 63:
			printf("Reserved: #%d\n",SF_VECTOR(framep));
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
		default:
			printf("User Defined Vector #%d\n",SF_VECTOR(framep));
			cpu_display_register(NULL);
			cpu_disasm((ADDRESS)context.pc,1);
			break;
	}

	if ((cpu_handler_flag == EXEC_DBUG) && step_over)
	{
			breakpoint_remove (step_over);
			step_over = 0;
	}

	return cpu_handler_flag;
}

/*************************************************/

#define REGREAD		(1)
#define REGWRITE	(0)

/*************************************************/
static void
irmd_sim (MCF5206_IMM *imm, char *reg, int regread, NATURAL32 value)
{
	int displayed = 0;
	int display_all = FALSE;

	if (reg == NULL)
	{
		pause(&displayed);
		display_all = TRUE;
	}

	if ((strcasecmp("SIMR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  SIMR:  %02X\n",imm->sim.SIMR);
			pause(&displayed);
		}
		else
			imm->sim.SIMR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR1",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR1:  %02X\n",imm->sim.ICR1);
			pause(&displayed);
		}
		else
			imm->sim.ICR1 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR2",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR2:  %02X\n",imm->sim.ICR2);
			pause(&displayed);
		}
		else
			imm->sim.ICR2 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR3",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR3:  %02X\n",imm->sim.ICR3);
			pause(&displayed);
		}
		else
			imm->sim.ICR3 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR4",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR4:  %02X\n",imm->sim.ICR4);
			pause(&displayed);
		}
		else
			imm->sim.ICR4 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR5",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR5:  %02X\n",imm->sim.ICR5);
			pause(&displayed);
		}
		else
			imm->sim.ICR5 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR6",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR6:  %02X\n",imm->sim.ICR6);
			pause(&displayed);
		}
		else
			imm->sim.ICR6 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR7",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR7:  %02X\n",imm->sim.ICR7);
			pause(&displayed);
		}
		else
			imm->sim.ICR7 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR8",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR8:  %02X\n",imm->sim.ICR8);
			pause(&displayed);
		}
		else
			imm->sim.ICR8 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR9",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR9:  %02X\n",imm->sim.ICR9);
			pause(&displayed);
		}
		else
			imm->sim.ICR9 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR10",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR10: %02X\n",imm->sim.ICR10);
			pause(&displayed);
		}
		else
			imm->sim.ICR10 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR11",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR11: %02X\n",imm->sim.ICR11);
			pause(&displayed);
		}
		else
			imm->sim.ICR11 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR12",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR12: %02X\n",imm->sim.ICR12);
			pause(&displayed);
		}
		else
			imm->sim.ICR12 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("ICR13",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  ICR13: %02X\n",imm->sim.ICR13);
			pause(&displayed);
		}
		else
			imm->sim.ICR13 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("IMR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  IMR:   %04X\n",imm->sim.IMR);
			pause(&displayed);
		}
		else
			imm->sim.IMR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("IPR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  IPR:   %04X\n",imm->sim.IPR);
			pause(&displayed);
		}
		else
			imm->sim.IPR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("RSR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  RSR:   %02X\n",imm->sim.RSR);
			pause(&displayed);
		}
		else
			imm->sim.RSR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("SYPCR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  SYPCR: %02X\n",imm->sim.SYPCR);
			pause(&displayed);
		}
		else
			imm->sim.SYPCR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("SWIVR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  SWIVR: %02X\n",imm->sim.SWIVR);
			pause(&displayed);
		}
		else
			imm->sim.SWIVR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("SWSR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  SWSR:  %02X\n",imm->sim.SWSR);
			pause(&displayed);
		}
		else
			imm->sim.SWSR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("PAR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  PAR:   %02X\n",imm->sim.PAR);
			pause(&displayed);
		}
		else
			imm->sim.PAR = value;

		if (!display_all)
			return;
	}

	if (!display_all)
		printf(INVREG,reg);
}

/*************************************************/
static void
irmd_gpio (MCF5206_IMM *imm, char *reg, int regread, NATURAL32 value)
{
	int display_all = FALSE;

	if (reg == NULL)
	{
		display_all = TRUE;
	}

	if ((strcasecmp("PPDDR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  PPDDR: %02X\n",imm->gpio.PPDDR);
		else
			imm->gpio.PPDDR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("PPDAT",reg) == 0) || display_all)
	{
		if (regread)
			printf("  PPDAT: %02X\n",imm->gpio.PPDAT);
		else
			imm->gpio.PPDAT = value;

		if (!display_all)
			return;
	}

	if (!display_all)
		printf(INVREG,reg);
}

/*************************************************/
static void
irmd_cs (MCF5206_IMM *imm, char *reg, int regread, NATURAL32 value)
{
	int displayed = 0;
	int display_all = FALSE;

	if (reg == NULL)
	{
		pause(&displayed);
		display_all = TRUE;
	}

	if ((strcasecmp("CSAR0",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSAR0: %04X\n",imm->cs.CSAR0);
			pause(&displayed);
		}
		else
			imm->cs.CSAR0 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSMR0",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSMR0: %08X\n",imm->cs.CSMR0);
			pause(&displayed);
		}
		else
			imm->cs.CSMR0 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSCR0",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSCR0: %04X\n",imm->cs.CSCR0);
			pause(&displayed);
		}
		else
			imm->cs.CSCR0 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSAR1",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSAR1: %04X\n",imm->cs.CSAR1);
			pause(&displayed);
		}
		else
			imm->cs.CSAR1 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSMR1",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSMR1: %08X\n",imm->cs.CSMR1);
			pause(&displayed);
		}
		else
			imm->cs.CSMR1 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSCR1",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSCR1: %04X\n",imm->cs.CSCR1);
			pause(&displayed);
		}
		else
			imm->cs.CSCR1 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSAR2",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSAR2: %04X\n",imm->cs.CSAR2);
			pause(&displayed);
		}
		else
			imm->cs.CSAR2 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSMR2",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSMR2: %08X\n",imm->cs.CSMR2);
			pause(&displayed);
		}
		else
			imm->cs.CSMR2 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSCR2",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSCR2: %04X\n",imm->cs.CSCR2);
			pause(&displayed);
		}
		else
			imm->cs.CSCR2 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSAR3",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSAR3: %04X\n",imm->cs.CSAR3);
			pause(&displayed);
		}
		else
			imm->cs.CSAR3 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSMR3",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSMR3: %08X\n",imm->cs.CSMR3);
			pause(&displayed);
		}
		else
			imm->cs.CSMR3 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSCR3",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSCR3: %04X\n",imm->cs.CSCR3);
			pause(&displayed);
		}
		else
			imm->cs.CSCR3 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSAR4",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSAR4: %04X\n",imm->cs.CSAR4);
			pause(&displayed);
		}
		else
			imm->cs.CSAR4 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSMR4",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSMR4: %08X\n",imm->cs.CSMR4);
			pause(&displayed);
		}
		else
			imm->cs.CSMR4 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSCR4",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSCR4: %04X\n",imm->cs.CSCR4);
			pause(&displayed);
		}
		else
			imm->cs.CSCR4 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSAR5",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSAR5: %04X\n",imm->cs.CSAR5);
			pause(&displayed);
		}
		else
			imm->cs.CSAR5 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSMR5",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSMR5: %08X\n",imm->cs.CSMR5);
			pause(&displayed);
		}
		else
			imm->cs.CSMR5 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSCR5",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSCR5: %04X\n",imm->cs.CSCR5);
			pause(&displayed);
		}
		else
			imm->cs.CSCR5 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSAR6",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSAR6: %04X\n",imm->cs.CSAR6);
			pause(&displayed);
		}
		else
			imm->cs.CSAR6 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSMR6",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSMR6: %08X\n",imm->cs.CSMR6);
			pause(&displayed);
		}
		else
			imm->cs.CSMR6 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSCR6",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSCR6: %04X\n",imm->cs.CSCR6);
			pause(&displayed);
		}
		else
			imm->cs.CSCR6 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSAR7",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSAR7: %04X\n",imm->cs.CSAR7);
			pause(&displayed);
		}
		else
			imm->cs.CSAR7 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSMR7",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSMR7: %08X\n",imm->cs.CSMR7);
			pause(&displayed);
		}
		else
			imm->cs.CSMR7 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("CSCR7",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  CSCR7: %04X\n",imm->cs.CSCR7);
			pause(&displayed);
		}
		else
			imm->cs.CSCR7 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("DMCR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DMCR:  %04X\n",imm->cs.DMCR);
			pause(&displayed);
		}
		else
			imm->cs.DMCR = value;

		if (!display_all)
			return;
	}

	if (!display_all)
		printf(INVREG,reg);
}

/*************************************************/
static void
irmd_dramc (MCF5206_IMM *imm, char *reg, int regread, NATURAL32 value)
{
	int displayed = 0;
	int display_all = FALSE;

	if (reg == NULL)
	{
		pause(&displayed);
		display_all = TRUE;
	}

	if ((strcasecmp("DCRR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DCRR:  %04X\n",imm->dramc.DCRR);
			pause(&displayed);
		}
		else
			imm->dramc.DCRR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("DCTR",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DCTR:  %04X\n",imm->dramc.DCTR);
			pause(&displayed);
		}
		else
			imm->dramc.DCTR = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("DCAR0",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DCAR0: %04X\n",imm->dramc.DCAR0);
			pause(&displayed);
		}
		else
			imm->dramc.DCAR0 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("DCMR0",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DCMR0: %08X\n",imm->dramc.DCMR0);
			pause(&displayed);
		}
		else
			imm->dramc.DCMR0 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("DCCR0",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DCCR0: %02X\n",imm->dramc.DCCR0);
			pause(&displayed);
		}
		else
			imm->dramc.DCCR0 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("DCAR1",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DCAR1: %04X\n",imm->dramc.DCAR1);
			pause(&displayed);
		}
		else
			imm->dramc.DCAR1 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("DCMR1",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DCMR1: %08X\n",imm->dramc.DCMR1);
			pause(&displayed);
		}
		else
			imm->dramc.DCMR1 = value;

		if (!display_all)
			return;
	}
	if ((strcasecmp("DCCR1",reg) == 0) || display_all)
	{
		if (regread)
		{
			printf("  DCCR1: %02X\n",imm->dramc.DCCR1);
			pause(&displayed);
		}
		else
			imm->dramc.DCCR1 = value;

		if (!display_all)
			return;
	}

	if (!display_all)
		printf(INVREG,reg);
}

/*************************************************/
static void
irmd_uart1 (MCF5206_IMM *imm, char *reg, int regread, NATURAL32 value)
{
	int display_all = FALSE;

	if (reg == NULL)
	{
		display_all = TRUE;
	}

	if ((strcasecmp("UMR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UMR:  %02X\n",imm->uart1.UMR);
		else
			imm->uart1.UMR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("USR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  USR:  %02X\n",imm->uart1.USR);
		else
			imm->uart1.USR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UCR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UCR:  %02X\n",imm->uart1.UCR);
		else
			imm->uart1.UCR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UBUF",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UBUF: %02X\n",imm->uart1.UBUF);
		else
			imm->uart1.UBUF = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UACR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UACR: %02X\n",imm->uart1.UACR);
		else
			imm->uart1.UACR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UIR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UIR:  %02X\n",imm->uart1.UIR);
		else
			imm->uart1.UIR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UBG1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UBG1: %02X\n",imm->uart1.UBG1);
		else
			imm->uart1.UBG1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UBG2",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UBG2: %02X\n",imm->uart1.UBG2);
		else
			imm->uart1.UBG2 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UIVR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UIVR: %02X\n",imm->uart1.UIVR);
		else
			imm->uart1.UIVR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UIP",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UIP:  %02X\n",imm->uart1.UIP);
		else
			/* Do Not Access */ ;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UOP1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UOP1: XX\n");
		else
			imm->uart1.UOP1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UOP0",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UOP0: XX\n");
		else
			imm->uart1.UOP0 = value;

		if (!display_all)
			return;
	}

	if (!display_all)
		printf(INVREG,reg);
}

/*************************************************/
static void
irmd_uart2 (MCF5206_IMM *imm, char *reg, int regread, NATURAL32 value)
{
	int display_all = FALSE;

	if (reg == NULL)
	{
		display_all = TRUE;
	}


	if ((strcasecmp("UMR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UMR:  %02X\n",imm->uart2.UMR);
		else
			imm->uart2.UMR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("USR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  USR:  %02X\n",imm->uart2.USR);
		else
			imm->uart2.USR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UCR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UCR:  %02X\n",imm->uart2.UCR);
		else
			imm->uart2.UCR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UBUF",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UBUF: %02X\n",imm->uart2.UBUF);
		else
			imm->uart2.UBUF = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UACR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UACR: %02X\n",imm->uart2.UACR);
		else
			imm->uart2.UACR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UIR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UIR:  %02X\n",imm->uart2.UIR);
		else
			imm->uart2.UIR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UBG1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UBG1: %02X\n",imm->uart2.UBG1);
		else
			imm->uart2.UBG1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UBG2",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UBG2: %02X\n",imm->uart2.UBG2);
		else
			imm->uart2.UBG2 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UIVR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UIVR: %02X\n",imm->uart2.UIVR);
		else
			imm->uart2.UIVR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UIP",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UIP:  %02X\n",imm->uart2.UIP);
		else
			/* Do Not Access */ ;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UOP1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UOP1: XX\n");
		else
			imm->uart2.UOP1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("UOP0",reg) == 0) || display_all)
	{
		if (regread)
			printf("  UOP0: XX\n");
		else
			imm->uart2.UOP0 = value;

		if (!display_all)
			return;
	}

	if (!display_all)
		printf(INVREG,reg);
}

/*************************************************/
static void
irmd_timer (MCF5206_IMM *imm, char *reg, int regread, NATURAL32 value)
{
	int display_all = FALSE;

	if (reg == NULL)
	{
		display_all = TRUE;
	}

	if ((strcasecmp("TMR1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TMR1: %04X\n",imm->timer.TMR1);
		else
			imm->timer.TMR1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TRR1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TRR1: %04X\n",imm->timer.TRR1);
		else
			imm->timer.TRR1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TCR1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TCR1: %04X\n",imm->timer.TCR1);
		else
			imm->timer.TCR1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TCN1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TCN1: %04X\n",imm->timer.TCN1);
		else
			imm->timer.TCN1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TER1",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TER1: %02X\n",imm->timer.TER1);
		else
			imm->timer.TER1 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TMR2",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TMR2: %04X\n",imm->timer.TMR2);
		else
			imm->timer.TMR2 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TRR2",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TRR2: %04X\n",imm->timer.TRR2);
		else
			imm->timer.TRR2 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TCR2",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TCR2: %04X\n",imm->timer.TCR2);
		else
			imm->timer.TCR2 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TCN2",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TCN2: %04X\n",imm->timer.TCN2);
		else
			imm->timer.TCN2 = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("TER2",reg) == 0) || display_all)
	{
		if (regread)
			printf("  TER2: %02X\n",imm->timer.TER2);
		else
			imm->timer.TER2 = value;

		if (!display_all)
			return;
	}

	if (!display_all)
		printf(INVREG,reg);
}

/*************************************************/
static void
irmd_mbus (MCF5206_IMM *imm, char *reg, int regread, NATURAL32 value)
{
	int display_all = FALSE;

	if (reg == NULL)
	{
		display_all = TRUE;
	}

	if ((strcasecmp("MADR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  MADR: %02X\n",imm->mbus.MADR);
		else
			imm->mbus.MADR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("MFDR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  MFDR: %02X\n",imm->mbus.MFDR);
		else
			imm->mbus.MFDR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("MBCR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  MBCR: %02X\n",imm->mbus.MBCR);
		else
			imm->mbus.MBCR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("MBSR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  MBSR: %02X\n",imm->mbus.MBSR);
		else
			imm->mbus.MBSR = value;

		if (!display_all)
			return;
	}

	if ((strcasecmp("MBDR",reg) == 0) || display_all)
	{
		if (regread)
			printf("  MBDR: %02X\n",imm->mbus.MBDR);
		else
			imm->mbus.MBDR = value;

		if (!display_all)
			return;
	}

	if (!display_all)
		printf(INVREG,reg);
}

/*************************************************/
static const struct module_t
{
	char	*module;
	void	(*func)(MCF5206_IMM *,char *, int, NATURAL32);
} MODULE[] =
{
	{"CS",		irmd_cs},
	{"DRAMC",	irmd_dramc},
	{"GPIO",	irmd_gpio},
	{"MBUS",	irmd_mbus},
	{"SIM",		irmd_sim},
	{"TIMER",	irmd_timer},
	{"UART1",	irmd_uart1},
	{"UART2",	irmd_uart2},
} ;

#define MODULE_SIZE (sizeof(MODULE)/sizeof(struct module_t))

/*************************************************/
void
ird (int argc, char **argv)
{
	/*
	 * Display Internal Memory Mapped register contents
	 */
	MCF5206_IMM *imm;
	char *rstr;
	char mstr[20];
	int mi;

	(void)argc;

	imm = mcf5206_get_immp();

	if (argv[1] == NULL)
	{
		printf("Internal Modules, MBAR, located at %#08X\n",imm);
		printf("Modules:  ");
		for (mi = 0; mi < MODULE_SIZE; ++mi)
		{
			printf("%s ",MODULE[mi].module);
		}
		printf("\n");
		return;
	}

	/*
	 * Pick out module name and point to register name
	 */
	mi = 0;
	rstr = argv[1];
	while (*rstr != '.')
	{
		mstr[mi++] = *rstr;
		if (*++rstr == '\0')
		{
			rstr = NULL;
			break;
		}
	}
	mstr[mi] = '\0';
	if (*rstr == '.')
		++rstr;

	/*
	 * Display the module contents
	 */
	for (mi = 0; mi < MODULE_SIZE; ++mi)
	{
		if (strcasecmp(MODULE[mi].module,mstr) == 0)
		{
			if (rstr == NULL)
			{
				printf("Module: %s\n",MODULE[mi].module);
			}
			MODULE[mi].func(imm, rstr, REGREAD, 0);
			return;
		}
	}

	printf(INVMOD,argv[1]);
}

/*************************************************/
void
irm (int argc, char **argv)
{
	/*
	 * Modify Internal Memory Mapped register contents
	 */
	NATURAL32 value;
	int success;
	MCF5206_IMM *imm;
	char *rstr;
	char mstr[20];
	int mi;

	(void)argc;

	imm = mcf5206_get_immp();

	/*
	 * Pick out module name and point to register name
	 */
	mi = 0;
	rstr = argv[1];
	while (*rstr != '.')
	{
		mstr[mi++] = *rstr;
		if (*++rstr == '\0')
		{
			rstr = NULL;
			break;
		}
	}
	mstr[mi] = '\0';
	if (*rstr == '.')
		++rstr;

	/*
	* Get the new value
	*/

	value = get_value(argv[2],&success,16);
	if(success == 0)
	{
		printf(INVALUE,argv[2]);
		return;
	}

	/*
	 * Modify the module contents
	 */
	for (mi = 0; mi < MODULE_SIZE; ++mi)
	{
		if (strcasecmp(MODULE[mi].module,mstr) == 0)
		{
			MODULE[mi].func(imm, rstr, REGWRITE, value);
			return;
		}
	}

	printf(INVMOD,argv[1]);
}

